/* Copyright (C) 2005-2019 Shugo Maeda <shugo@ruby-lang.org> and Charlie Savage
   <cfis@savagexi.com> Please see the LICENSE file for copyright and
   distribution information */

#include "bf_measurement.h"

VALUE mMeasure;
VALUE cRpMeasurement;

prof_measurer_t *prof_measurer_wall_process(void);
void bf_init_measure(void);

prof_measurer_t *prof_get_measurer(void) { return prof_measurer_wall_process(); };

struct four_doubles_t prof_measure(prof_measurer_t *measurer, rb_trace_arg_t *trace_arg)
{
    struct four_doubles_t ret;
    struct four_doubles_t measurement = measurer->measure(trace_arg);

    ret.wt = measurement.wt * measurer->multipliers.wt;
    ret.pr = measurement.pr * measurer->multipliers.pr;
    ret.al = measurement.al * measurer->multipliers.al;
    ret.mm = measurement.mm * measurer->multipliers.mm;
    return ret;
}

/* =======  prof_measurement_t   ========*/
prof_measurement_t *prof_measurement_create(void)
{
    prof_measurement_t *result = ALLOC(prof_measurement_t);
    result->wt_total_time = 0;
    result->wt_self_time = 0;
    result->wt_wait_time = 0;

    result->pr_total_time = 0;
    result->pr_self_time = 0;
    result->pr_wait_time = 0;

    result->al_total_time = 0;
    result->al_self_time = 0;
    result->al_wait_time = 0;

    result->mm_self_time = 0;
    result->mm_wait_time = 0;

    result->called = 0;
    result->object = Qnil;
    return result;
}

void prof_measurement_mark(void *data)
{
    if (!data)
    {
        return;
    }

    prof_measurement_t *measurement_data = (prof_measurement_t *)data;

    if (measurement_data->object != Qnil)
    {
        rb_gc_mark(measurement_data->object);
    }
}

static void prof_measurement_ruby_gc_free(void *data)
{
    if (data)
    {
        // Measurements are freed by their owning object (call info or method)
        prof_measurement_t *measurement = (prof_measurement_t *)data;
        measurement->object = Qnil;
    }
}

void prof_measurement_free(prof_measurement_t *measurement)
{
    /* Has this measurement object been accessed by Ruby?  If
       yes clean it up so to avoid a segmentation fault. */
    if (measurement->object != Qnil)
    {
        RTYPEDDATA(measurement->object)->data = NULL;
        measurement->object = Qnil;
    }

    xfree(measurement);
}

size_t prof_measurement_size(const void *data) { return sizeof(prof_measurement_t); }

static const rb_data_type_t measurement_type = {.wrap_struct_name = "Measurement",
                                                .function =
                                                    {
                                                        .dmark = prof_measurement_mark,
                                                        .dfree = prof_measurement_ruby_gc_free,
                                                        .dsize = prof_measurement_size,
                                                    },
                                                .data = NULL,
                                                .flags = RUBY_TYPED_FREE_IMMEDIATELY};

VALUE prof_measurement_wrap(prof_measurement_t *measurement)
{
    if (measurement->object == Qnil)
    {
        measurement->object = TypedData_Wrap_Struct(cRpMeasurement, &measurement_type, measurement);
    }
    return measurement->object;
}

static VALUE prof_measurement_allocate(VALUE klass)
{
    prof_measurement_t *measurement = prof_measurement_create();
    measurement->object = prof_measurement_wrap(measurement);
    return measurement->object;
}

prof_measurement_t *prof_get_measurement(VALUE self)
{
    /* Can't use Data_Get_Struct because that triggers the event hook
       ending up in endless recursion. */
    prof_measurement_t *result = RTYPEDDATA_DATA(self);

    if (!result)
    {
        rb_raise(rb_eRuntimeError, "This Blackfire::Measurement instance has already been freed, "
                                   "likely because its profile has been freed.");
    }

    return result;
}

/* call-seq:
   wt_total_time -> float

   Returns the total amount of time spent in this method and its children. */
static VALUE prof_measurement_wt_total_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->wt_total_time);
}

/* call-seq:
   wt_self_time -> float

   Returns the total amount of time spent in this method. */
static VALUE prof_measurement_wt_self_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->wt_self_time);
}

/* call-seq:
   wt_wait_time -> float

   Returns the total amount of time this method waited for other threads. */
static VALUE prof_measurement_wt_wait_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->wt_wait_time);
}

/* call-seq:
   pr_total_time -> float

   Returns the total amount of time spent in this method and its children. */
static VALUE prof_measurement_pr_total_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->pr_total_time);
}

/* call-seq:
   pr_self_time -> float

   Returns the total amount of time spent in this method. */
static VALUE prof_measurement_pr_self_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->pr_self_time);
}

/* call-seq:
   pr_wait_time -> float

   Returns the total amount of time this method waited for other threads. */
static VALUE prof_measurement_pr_wait_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->pr_wait_time);
}

/* call-seq:
   al_total_time -> float

   Returns the total amount of time spent in this method and its children. */
static VALUE prof_measurement_al_total_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->al_total_time);
}

/* call-seq:
   al_self_time -> float

   Returns the total amount of time spent in this method. */
static VALUE prof_measurement_al_self_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->al_self_time);
}

/* call-seq:
   al_wait_time -> float

   Returns the total amount of time this method waited for other threads. */
static VALUE prof_measurement_al_wait_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->al_wait_time);
}

/* call-seq:
   mm_total_time -> float

   Returns the total amount of time spent in this method and its children. */
static VALUE prof_measurement_mm_total_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->mm_self_time);
}

/* call-seq:
   mm_self_time -> float

   Returns the total amount of time spent in this method. */
static VALUE prof_measurement_mm_self_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->mm_self_time);
}

/* call-seq:
   mm_wait_time -> float

   Returns the total amount of time this method waited for other threads. */
static VALUE prof_measurement_mm_wait_time(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return rb_float_new(result->mm_wait_time);
}

/* call-seq:
   called -> int

   Returns the total amount of times this method was called. */
static VALUE prof_measurement_called(VALUE self)
{
    prof_measurement_t *result = prof_get_measurement(self);
    return INT2NUM(result->called);
}

/* call-seq:
   called=n -> n

   Sets the call count to n. */
static VALUE prof_measurement_set_called(VALUE self, VALUE called)
{
    prof_measurement_t *result = prof_get_measurement(self);
    result->called = NUM2INT(called);
    return called;
}

/* :nodoc: */
static VALUE prof_measurement_dump(VALUE self)
{
    prof_measurement_t *measurement_data = prof_get_measurement(self);
    VALUE result = rb_hash_new();

    rb_hash_aset(result, ID2SYM(rb_intern("wall_time_total_time")), rb_float_new(measurement_data->wt_total_time));
    rb_hash_aset(result, ID2SYM(rb_intern("wall_time_self_time")), rb_float_new(measurement_data->wt_self_time));
    rb_hash_aset(result, ID2SYM(rb_intern("wall_time_wait_time")), rb_float_new(measurement_data->wt_wait_time));

    rb_hash_aset(result, ID2SYM(rb_intern("process_total_time")), rb_float_new(measurement_data->pr_total_time));
    rb_hash_aset(result, ID2SYM(rb_intern("process_self_time")), rb_float_new(measurement_data->pr_self_time));
    rb_hash_aset(result, ID2SYM(rb_intern("process_wait_time")), rb_float_new(measurement_data->pr_wait_time));

    rb_hash_aset(result, ID2SYM(rb_intern("allocation_total_time")), rb_float_new(measurement_data->al_total_time));
    rb_hash_aset(result, ID2SYM(rb_intern("allocation_self_time")), rb_float_new(measurement_data->al_self_time));
    rb_hash_aset(result, ID2SYM(rb_intern("allocation_wait_time")), rb_float_new(measurement_data->al_wait_time));

    rb_hash_aset(result, ID2SYM(rb_intern("memory_total_time")), rb_float_new(measurement_data->mm_self_time));
    rb_hash_aset(result, ID2SYM(rb_intern("memory_self_time")), rb_float_new(measurement_data->mm_self_time));
    rb_hash_aset(result, ID2SYM(rb_intern("memory_wait_time")), rb_float_new(measurement_data->mm_wait_time));

    rb_hash_aset(result, ID2SYM(rb_intern("called")), INT2FIX(measurement_data->called));

    return result;
}

/* :nodoc: */
static VALUE prof_measurement_load(VALUE self, VALUE data)
{
    prof_measurement_t *measurement = prof_get_measurement(self);
    measurement->object = self;

    measurement->wt_total_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("wall_time_total_time"))));
    measurement->wt_self_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("wall_time_self_time"))));
    measurement->wt_wait_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("wall_time_wait_time"))));

    measurement->pr_total_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("process_total_time"))));
    measurement->pr_self_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("process_self_time"))));
    measurement->pr_wait_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("process_wait_time"))));

    measurement->al_total_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("allocation_total_time"))));
    measurement->al_self_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("allocation_self_time"))));
    measurement->al_wait_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("allocation_wait_time"))));

    measurement->mm_self_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("memory_total_time"))));
    measurement->mm_self_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("memory_self_time"))));
    measurement->mm_wait_time = rb_num2dbl(rb_hash_aref(data, ID2SYM(rb_intern("memory_wait_time"))));

    measurement->called = FIX2INT(rb_hash_aref(data, ID2SYM(rb_intern("called"))));

    return data;
}

void rp_init_measure()
{
    mMeasure = rb_define_module_under(mProf, "Measure");
    bf_init_measure();

    cRpMeasurement = rb_define_class_under(mProf, "Measurement", rb_cObject);
    rb_undef_method(CLASS_OF(cRpMeasurement), "new");
    rb_define_alloc_func(cRpMeasurement, prof_measurement_allocate);

    rb_define_method(cRpMeasurement, "called", prof_measurement_called, 0);
    rb_define_method(cRpMeasurement, "called=", prof_measurement_set_called, 1);

    rb_define_method(cRpMeasurement, "wall_time_total_time", prof_measurement_wt_total_time, 0);
    rb_define_method(cRpMeasurement, "wall_time_self_time", prof_measurement_wt_self_time, 0);
    rb_define_method(cRpMeasurement, "wall_time_wait_time", prof_measurement_wt_wait_time, 0);

    rb_define_method(cRpMeasurement, "process_total_time", prof_measurement_pr_total_time, 0);
    rb_define_method(cRpMeasurement, "process_self_time", prof_measurement_pr_self_time, 0);
    rb_define_method(cRpMeasurement, "process_wait_time", prof_measurement_pr_wait_time, 0);

    rb_define_method(cRpMeasurement, "allocation_total_time", prof_measurement_al_total_time, 0);
    rb_define_method(cRpMeasurement, "allocation_self_time", prof_measurement_al_self_time, 0);
    rb_define_method(cRpMeasurement, "allocation_wait_time", prof_measurement_al_wait_time, 0);

    rb_define_method(cRpMeasurement, "memory_total_time", prof_measurement_mm_total_time, 0);
    rb_define_method(cRpMeasurement, "memory_self_time", prof_measurement_mm_self_time, 0);
    rb_define_method(cRpMeasurement, "memory_wait_time", prof_measurement_mm_wait_time, 0);

    rb_define_method(cRpMeasurement, "_dump_data", prof_measurement_dump, 0);
    rb_define_method(cRpMeasurement, "_load_data", prof_measurement_load, 1);
}
