/* Copyright (C) 2005-2019 Shugo Maeda <shugo@ruby-lang.org> and Charlie Savage
   <cfis@savagexi.com> Please see the LICENSE file for copyright and
   distribution information */

/* Document-class: Blackfire::Profile

   The Profile class represents a single profiling run and provides the main API
   for using blackfire. After creating a Profile instance, start profiling code
   by calling the Profile#start method. To finish profiling, call Profile#stop.
   Once profiling is completed, the Profile instance contains the results.

   profile = Blackfire::Profile.new
   profile.start
   ...
   result = profile.stop

   Alternatively, you can use the block syntax:

   profile = Blackfire::Profile.profile do
   ...
   end
*/

#include "bf_profile.h"
#include "bf_allocation.h"
#include "bf_call_tree.h"
#include "bf_call_trees.h"
#include "bf_method.h"
#include <assert.h>

#define MULTIPLIER_FOR_TIME 1000000
#define SPECIAL_CHARS_FOR_TIMELINE_FORMAT 9

VALUE cProfile;
VALUE get_fiber(prof_profile_t *profile) { return rb_thread_current(); }

/* static const char* get_event_name(rb_event_flag_t event) */
/* { */
/*     switch (event) { */
/*     case RUBY_EVENT_LINE: */
/*         return "line"; */
/*     case RUBY_EVENT_CLASS: */
/*         return "class"; */
/*     case RUBY_EVENT_END: */
/*         return "end"; */
/*     case RUBY_EVENT_CALL: */
/*         return "call"; */
/*     case RUBY_EVENT_RETURN: */
/*         return "return"; */
/*     case RUBY_EVENT_B_CALL: */
/*         return "b-call"; */
/*     case RUBY_EVENT_B_RETURN: */
/*         return "b-return"; */
/*     case RUBY_EVENT_C_CALL: */
/*         return "c-call"; */
/*     case RUBY_EVENT_C_RETURN: */
/*         return "c-return"; */
/*     case RUBY_EVENT_THREAD_BEGIN: */
/*         return "thread-begin"; */
/*     case RUBY_EVENT_THREAD_END: */
/*         return "thread-end"; */
/*     case RUBY_EVENT_FIBER_SWITCH: */
/*         return "fiber-switch"; */
/*     case RUBY_EVENT_RAISE: */
/*         return "raise"; */
/*     case RUBY_INTERNAL_EVENT_NEWOBJ: */
/*         return "newobj"; */
/*     default: */
/*         return "unknown"; */
/*     } */
/* } */

thread_data_t *check_fiber(prof_profile_t *profile, struct four_doubles_t measurement)
{
    thread_data_t *result = NULL;

    /* Get the current thread and fiber information. */
    VALUE fiber = get_fiber(profile);

    /* We need to switch the profiling context if we either had none before,
       we don't merge fibers and the fiber ids differ, or the thread ids differ.
     */
    if (profile->last_thread_data->fiber != fiber)
    {
        result = threads_table_lookup(profile, fiber);
        if (!result)
        {
            result = threads_table_insert(profile, fiber);
        }
        switch_thread(profile, result, measurement);
    }
    else
    {
        result = profile->last_thread_data;
    }
    return result;
}

static int excludes_method(st_data_t key, prof_profile_t *profile)
{
    return (profile->exclude_methods_tbl && method_table_lookup(profile->exclude_methods_tbl, key) != NULL);
}

static prof_method_t *create_method(VALUE profile, st_data_t key, VALUE klass, VALUE msym, VALUE source_file, int source_line)
{
    prof_method_t *result = prof_method_create(profile, klass, msym, source_file, source_line);

    prof_profile_t *profile_t = prof_get_profile(profile);
    method_table_insert(profile_t->last_thread_data->method_table, result->key, result);

    return result;
}

static prof_method_t *check_parent_method(VALUE profile, thread_data_t *thread_data)
{
    VALUE msym = ID2SYM(rb_intern("_inserted_parent_"));
    st_data_t key = method_key(cProfile, msym);

    prof_method_t *result = method_table_lookup(thread_data->method_table, key);

    if (!result)
    {
        result = create_method(profile, key, cProfile, msym, Qnil, 0);
    }

    return result;
}

prof_method_t *check_method(VALUE profile, rb_trace_arg_t *trace_arg, rb_event_flag_t event, thread_data_t *thread_data)
{
    VALUE klass = rb_tracearg_defined_class(trace_arg);

    /* Special case - skip any methods from the mProf
       module or cProfile class since they clutter
       the results but aren't important to them results. */
    if (klass == cProfile)
    {
        return NULL;
    }

#ifdef HAVE_RB_TRACEARG_CALLEE_ID
    VALUE msym = rb_tracearg_callee_id(trace_arg);
#else
    VALUE msym = rb_tracearg_method_id(trace_arg);
#endif

    st_data_t key = method_key(klass, msym);

    prof_profile_t *profile_t = prof_get_profile(profile);
    if (excludes_method(key, profile_t))
    {
        return NULL;
    }

    prof_method_t *result = method_table_lookup(thread_data->method_table, key);

    if (!result)
    {
        VALUE source_file = (event != RUBY_EVENT_C_CALL ? rb_tracearg_path(trace_arg) : Qnil);
        int source_line = (event != RUBY_EVENT_C_CALL ? FIX2INT(rb_tracearg_lineno(trace_arg)) : 0);
        result = create_method(profile, key, klass, msym, source_file, source_line);
    }

    return result;
}

static char *escape_method_name(char *method_name)
{
    char *ret = malloc(strlen(method_name) + SPECIAL_CHARS_FOR_TIMELINE_FORMAT);
    if (ret == NULL)
    {
        return NULL;
    }

    long unsigned int current_position = 0;
    long unsigned int original_position = 0;
    char current_char;

    for (original_position = 0; original_position <= strlen(method_name); original_position++)
    {
        current_char = method_name[original_position];

        switch (current_char)
        {
        case '?':
        {
            ret[current_position++] = '%';
            ret[current_position++] = '3';
            ret[current_position++] = 'F';
            break;
        }
        case '!':
        {
            ret[current_position++] = '%';
            ret[current_position++] = '2';
            ret[current_position++] = '1';
            break;
        }
        default:
        {
            ret[current_position++] = current_char;
            break;
        }
        }
    }
    return ret;
}

static bool parse_method_name_for_timeline(prof_profile_t *profile, char *c_method, VALUE ruby_method, long rule_id)
{
    VALUE str;
    char *c_str = NULL;
    int ret = 0;
    str = rb_sprintf("%" PRIsVALUE, rb_ary_entry(profile->timeline_rules, rule_id));
    c_str = StringValueCStr(str);
    if (c_str[strlen(c_str) - 1] == '\n')
    {
        c_str[strlen(c_str) - 1] = '\0';
    }

    char *fqmn = NULL;
    if (RB_TYPE_P(ruby_method, T_STRING))
    {
        VALUE str;
        str = rb_sprintf("%" PRIsVALUE, ruby_method);
        fqmn = StringValueCStr(str);
    }
    else
    {
        fqmn = c_method;
    }

    if (c_str[0] == '=')
    {
        c_str++;
        ret = strcmp(fqmn, c_str);
        if (ret == 0)
        {
            return true;
        }
    }
    if (c_str[0] == '^')
    {
        c_str++;
        if (strncmp(fqmn, c_str, strlen(c_str)) == 0)
        {
            return true;
        }
    }
    return false;
}

static bool remove_internal_methods(char *c_method)
{
    char blackfire_string[] = "Blackfire::";

    if (strncmp(c_method, blackfire_string, strlen(blackfire_string)) == 0)
    {
        return true;
    }

    return false;
}

static char *generate_timespan_entry(prof_profile_t *profile, rb_trace_arg_t *trace_arg, struct four_doubles_t measurement)
{
    struct prof_timeline_item_t *tl = (prof_timeline_item_t *)malloc(sizeof(prof_timeline_item_t));
    rb_event_flag_t event = rb_tracearg_event_flag(trace_arg);

#ifdef HAVE_RB_TRACEARG_CALLEE_ID
    VALUE msym = rb_tracearg_callee_id(trace_arg);
#else
    VALUE msym = rb_tracearg_method_id(trace_arg);
#endif

    unsigned int klass_flags;
    VALUE klass = rb_tracearg_defined_class(trace_arg);
    VALUE resolved_klass = resolve_klass(klass, &klass_flags);
    const char *class_name = "";

    if (resolved_klass != Qnil)
    {
        class_name = rb_class2name(resolved_klass);
    }

    const char *method_name_char = (msym != Qnil ? rb_id2name(SYM2ID(msym)) : "");
    unsigned long fqmn_size = strlen(method_name_char) + strlen(class_name) + 2;
    char *fqmn = malloc(fqmn_size);
    if (fqmn == NULL)
    {
        goto ERROR;
    }

    unsigned long ret = snprintf(fqmn, fqmn_size, "%s#%s", class_name, method_name_char);
    if (ret > fqmn_size)
    {
        goto ERROR;
    }
    if (remove_internal_methods(fqmn))
    {
        goto ERROR;
    }

    if (profile->first_threshold == true)
    {
        profile->threshold_initial_time = (long long)(measurement.wt * MULTIPLIER_FOR_TIME);
        profile->first_threshold = false;
    }
    if (((long long)(measurement.wt * MULTIPLIER_FOR_TIME)) < profile->threshold_initial_time)
    {
        profile->threshold_initial_time = (long long)(measurement.wt * MULTIPLIER_FOR_TIME);
    }

    long long wt = (long long)((measurement.wt * MULTIPLIER_FOR_TIME) - profile->threshold_initial_time);
    long long pr = (long long)(measurement.pr * MULTIPLIER_FOR_TIME);
    long long mm = (long long)measurement.mm;
    long long al = (long long)measurement.al;

    switch (event)
    {
    case RUBY_EVENT_CLASS:
    case RUBY_EVENT_END:
    case RUBY_EVENT_THREAD_BEGIN:
    case RUBY_EVENT_THREAD_END:
    case RUBY_EVENT_FIBER_SWITCH:
    case RUBY_EVENT_RAISE:
    case RUBY_INTERNAL_EVENT_NEWOBJ:
    {
        return NULL;
        break;
    }
    case RUBY_EVENT_LINE:
    {
        if (profile->last_event == event)
        {
            return NULL;
        }

        if (profile->current_method != NULL)
        {
            if (strcmp(profile->current_method->fqmn, fqmn) == 0)
            {
                return NULL;
            }
        }

        tl->caller = profile->current_method;
        tl->fqmn = fqmn;

        tl->wt = wt;
        tl->pr = pr;
        tl->al = al;
        tl->mm = mm;

        profile->current_method = tl;
        break;
    }
    case RUBY_EVENT_CALL:
    case RUBY_EVENT_C_CALL:
    {
        tl->caller = profile->current_method;
        tl->fqmn = fqmn;

        tl->wt = wt;
        tl->pr = pr;
        tl->al = al;
        tl->mm = mm;

        profile->current_method = tl;
        break;
    }
    case RUBY_EVENT_RETURN:
    case RUBY_EVENT_C_RETURN:
    {
        if (profile->current_method != NULL)
        {
            if (profile->current_method->caller != NULL)
            {
                /* if wall time < threshold because if >= will be included for sure */
                if (wt < profile->timespan_threshold)
                {
                    bool should_include = false;
                    /* check timespan rules for each method */
                    for (int i = 0; i < profile->timeline_rules_count; i++)
                    {
                        should_include = parse_method_name_for_timeline(profile, fqmn, Qnil, i);
                        if (should_include)
                        {
                            break;
                        }
                    }

                    if (!should_include)
                    {
                        goto ERROR;
                    }
                }

                char *first_part = malloc(START_MAX_LEN);
                ret = snprintf(first_part, START_MAX_LEN, "Threshold-%d-Start: //%lld %lld %lld %lld", profile->current_threshold, profile->current_method->wt,
                               profile->current_method->pr, profile->current_method->al, profile->current_method->mm);
                if (ret > START_MAX_LEN)
                {
                    goto ERROR;
                }

                char *ret_fqmn = escape_method_name(fqmn);
                char *ret_caller_fqmn = escape_method_name(profile->current_method->caller->fqmn);
                unsigned long end_max_len = START_MAX_LEN + strlen(ret_fqmn) + strlen(ret_caller_fqmn);
                char *last_part = malloc(end_max_len);
                ret = snprintf(last_part, end_max_len, "Threshold-%d-End: %s==>%s//%lld %lld %lld %lld", profile->current_threshold, ret_caller_fqmn, ret_fqmn,
                               wt, pr, al, mm);
                if (ret > end_max_len)
                {
                    goto ERROR;
                }

                unsigned long out_size = strlen(first_part) + strlen(last_part) + 3;
                char *line = malloc(out_size);
                if (line == NULL)
                {
                    goto ERROR;
                }

                ret = snprintf(line, out_size, "%s\n%s\n", first_part, last_part);
                if (ret > out_size)
                {
                    free(first_part);
                    free(last_part);
                    goto ERROR;
                }

                free(first_part);
                free(last_part);

                tl = profile->current_method;
                profile->current_method = tl->caller;
                free(tl);

                profile->current_threshold++;
                free(fqmn);
                profile->last_event = event;
                return line;

            ERROR:
                if (profile->current_method != NULL)
                {
                    if (profile->current_method->caller != NULL)
                    {
                        profile->current_method = profile->current_method->caller;
                    }
                }

                free(fqmn);
            }
        }
        break;
    }
    }
    profile->last_event = event;
    return NULL;
}

static void prof_event_hook(VALUE trace_point, void *data)
{
    VALUE profile = (VALUE)data;
    prof_profile_t *profile_t = prof_get_profile(profile);

    rb_trace_arg_t *trace_arg = rb_tracearg_from_tracepoint(trace_point);
    struct four_doubles_t measurement = prof_measure(profile_t->measurer, trace_arg);
    rb_event_flag_t event = rb_tracearg_event_flag(trace_arg);
    VALUE self = rb_tracearg_self(trace_arg);

    /* Special case - skip any methods from the mProf
       module since they clutter the results but aren't important to them
       results.
     */
    if (self == mProf)
    {
        return;
    }

    thread_data_t *thread_data = check_fiber(profile_t, measurement);
    if (!thread_data->trace)
    {
        return;
    }

    /* add calls to timeline */
    if (profile_t->timeline_defined == true)
    {
        if (profile_t->paused == false)
        {
            char *data_line = generate_timespan_entry(profile_t, trace_arg, measurement);
            if (data_line != NULL)
            {
                rb_str_cat2(profile_t->timelines, data_line);
            }
            free(data_line);
        }
    }

    switch (event)
    {
    case RUBY_EVENT_LINE:
    {
        prof_frame_t *frame = prof_frame_current(thread_data->stack);

        if (!frame)
        {
            prof_method_t *method = check_method(profile, trace_arg, event, thread_data);

            if (!method)
            {
                break;
            }

            prof_call_tree_t *call_tree = prof_call_tree_create(method, NULL, method->source_file, method->source_line);
            prof_add_call_tree(method->call_trees, call_tree);

            if (thread_data->call_tree)
            {
                prof_call_tree_add_parent(thread_data->call_tree, call_tree);
                frame = prof_frame_unshift(thread_data->stack, call_tree, thread_data->call_tree, measurement);
            }
            else
            {
                frame = prof_frame_push(thread_data->stack, call_tree, measurement, RTEST(profile_t->paused));
            }

            thread_data->call_tree = call_tree;
        }

        frame->source_file = rb_tracearg_path(trace_arg);
        frame->source_line = FIX2INT(rb_tracearg_lineno(trace_arg));

        break;
    }
    case RUBY_EVENT_CALL:
    case RUBY_EVENT_C_CALL:
    {
        prof_method_t *method = check_method(profile, trace_arg, event, thread_data);

        if (!method)
        {
            break;
        }

        prof_frame_t *frame = prof_frame_current(thread_data->stack);
        prof_call_tree_t *parent_call_tree = NULL;
        prof_call_tree_t *call_tree = NULL;

        // Frame can be NULL if we are switching from one fiber to another (see
        // FiberTest#fiber_test)
        if (frame)
        {
            parent_call_tree = frame->call_tree;
            call_tree = call_tree_table_lookup(parent_call_tree->children, method->key);
        }
        else if (!frame && thread_data->call_tree)
        {
            // There is no current parent - likely we have returned out of the
            // highest level method we have profiled so far. This can happen
            // with enumerators (see fiber_test.rb). So create a new dummy
            // parent.
            prof_method_t *parent_method = check_parent_method(profile, thread_data);
            parent_call_tree = prof_call_tree_create(parent_method, NULL, Qnil, 0);
            prof_add_call_tree(parent_method->call_trees, parent_call_tree);
            prof_call_tree_add_parent(thread_data->call_tree, parent_call_tree);
            frame = prof_frame_unshift(thread_data->stack, parent_call_tree, thread_data->call_tree, measurement);
            thread_data->call_tree = parent_call_tree;
        }

        if (!call_tree)
        {
            // This call info does not yet exist.  So create it and add it to
            // previous CallTree's children and the current method.
            call_tree = prof_call_tree_create(method, parent_call_tree, frame ? frame->source_file : Qnil, frame ? frame->source_line : 0);
            prof_add_call_tree(method->call_trees, call_tree);
            if (parent_call_tree)
            {
                prof_call_tree_add_child(parent_call_tree, call_tree);
            }
        }

        if (!thread_data->call_tree)
        {
            thread_data->call_tree = call_tree;
        }

        // Push a new frame onto the stack for a new c-call or ruby call (into a
        // method)
        prof_frame_t *next_frame = prof_frame_push(thread_data->stack, call_tree, measurement, RTEST(profile_t->paused));
        next_frame->source_file = method->source_file;
        next_frame->source_line = method->source_line;
        break;
    }
    case RUBY_EVENT_RETURN:
    case RUBY_EVENT_C_RETURN:
    {
        // We need to check for excluded methods so that we don't pop them off
        // the stack
        prof_method_t *method = check_method(profile, trace_arg, event, thread_data);

        if (!method)
        {
            break;
        }

        prof_frame_pop(thread_data->stack, measurement, profile_t->measurer);
        break;
    }
    case RUBY_INTERNAL_EVENT_NEWOBJ:
    {
        /* We want to assign the allocations lexically, not the execution
           context (otherwise all allocations will show up under Class#new */
        int source_line = FIX2INT(rb_tracearg_lineno(trace_arg));
        VALUE source_file = rb_tracearg_path(trace_arg);

        prof_method_t *method = prof_find_method(thread_data->stack, source_file, source_line);
        if (method)
        {
            prof_allocate_increment(method, trace_arg);
        }

        break;
    }
    }
}

void prof_install_hook(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);

    VALUE event_tracepoint =
        rb_tracepoint_new(Qnil, RUBY_EVENT_CALL | RUBY_EVENT_RETURN | RUBY_EVENT_C_CALL | RUBY_EVENT_C_RETURN | RUBY_EVENT_LINE, prof_event_hook, (void *)self);
    rb_ary_push(profile->tracepoints, event_tracepoint);

    VALUE allocation_tracepoint = rb_tracepoint_new(Qnil, RUBY_INTERNAL_EVENT_NEWOBJ, prof_event_hook, (void *)self);
    rb_ary_push(profile->tracepoints, allocation_tracepoint);

    for (int i = 0; i < RARRAY_LEN(profile->tracepoints); i++)
    {
        rb_tracepoint_enable(rb_ary_entry(profile->tracepoints, i));
    }
}

void prof_remove_hook(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);

    for (int i = 0; i < RARRAY_LEN(profile->tracepoints); i++)
    {
        rb_tracepoint_disable(rb_ary_entry(profile->tracepoints, i));
    }
    rb_ary_clear(profile->tracepoints);
}

prof_profile_t *prof_get_profile(VALUE self)
{
    /* Can't use Data_Get_Struct because that triggers the event hook
       ending up in endless recursion. */
    return RTYPEDDATA_DATA(self);
}

static int collect_threads(st_data_t key, st_data_t value, st_data_t result)
{
    thread_data_t *thread_data = (thread_data_t *)value;
    if (thread_data->trace)
    {
        VALUE threads_array = (VALUE)result;
        rb_ary_push(threads_array, prof_thread_wrap(thread_data));
    }
    return ST_CONTINUE;
}

/* ========  Profile Class ====== */
static int mark_threads(st_data_t key, st_data_t value, st_data_t result)
{
    thread_data_t *thread = (thread_data_t *)value;
    prof_thread_mark(thread);
    return ST_CONTINUE;
}

static int prof_profile_mark_methods(st_data_t key, st_data_t value, st_data_t result)
{
    prof_method_t *method = (prof_method_t *)value;
    prof_method_mark(method);
    return ST_CONTINUE;
}

static void prof_profile_mark(void *data)
{
    prof_profile_t *profile = (prof_profile_t *)data;
    rb_gc_mark(profile->tracepoints);
    rb_gc_mark(profile->running);
    rb_gc_mark(profile->paused);
    rb_gc_mark(profile->timelines);
    rb_gc_mark(profile->headers);
    rb_gc_mark(profile->measurements);
    rb_gc_mark(profile->timeline_rules);

    // If GC stress is true (useful for debugging), when threads_table_create is
    // called in the allocate method Ruby will immediately call this mark
    // method. Thus the threads_tbl will be NULL.
    if (profile->threads_tbl)
    {
        rb_st_foreach(profile->threads_tbl, mark_threads, 0);
    }

    if (profile->exclude_methods_tbl)
    {
        rb_st_foreach(profile->exclude_methods_tbl, prof_profile_mark_methods, 0);
    }
}

/* Freeing the profile creates a cascade of freeing. It frees its threads table,
   which frees each thread and its associated call treee and methods. */
static void prof_profile_ruby_gc_free(void *data)
{
    prof_profile_t *profile = (prof_profile_t *)data;
    profile->last_thread_data = NULL;

    threads_table_free(profile->threads_tbl);
    profile->threads_tbl = NULL;

    if (profile->exclude_threads_tbl)
    {
        rb_st_free_table(profile->exclude_threads_tbl);
        profile->exclude_threads_tbl = NULL;
    }

    if (profile->include_threads_tbl)
    {
        rb_st_free_table(profile->include_threads_tbl);
        profile->include_threads_tbl = NULL;
    }

    /* This table owns the excluded sentinels for now. */
    method_table_free(profile->exclude_methods_tbl);
    profile->exclude_methods_tbl = NULL;

    xfree(profile->measurer);
    profile->measurer = NULL;

    xfree(profile);
}

size_t prof_profile_size(const void *data) { return sizeof(prof_profile_t); }

static const rb_data_type_t profile_type = {.wrap_struct_name = "Profile",
                                            .function =
                                                {
                                                    .dmark = prof_profile_mark,
                                                    .dfree = prof_profile_ruby_gc_free,
                                                    .dsize = prof_profile_size,
                                                },
                                            .data = NULL,
                                            .flags = RUBY_TYPED_FREE_IMMEDIATELY};

static VALUE prof_allocate(VALUE klass)
{
    VALUE result;
    prof_profile_t *profile;
    result = TypedData_Make_Struct(klass, prof_profile_t, &profile_type, profile);
    profile->threads_tbl = threads_table_create();
    profile->exclude_threads_tbl = NULL;
    profile->include_threads_tbl = NULL;
    profile->running = Qfalse;
    profile->allow_exceptions = false;
    profile->merge_fibers = false;
    profile->exclude_methods_tbl = method_table_create();
    profile->running = Qfalse;
    profile->tracepoints = rb_ary_new();
    return result;
}

static void prof_exclude_common_methods(VALUE profile) { rb_funcall(profile, rb_intern("exclude_common_methods!"), 0); }

static int pop_frames(VALUE key, st_data_t value, st_data_t data)
{
    thread_data_t *thread_data = (thread_data_t *)value;
    prof_profile_t *profile = (prof_profile_t *)data;
    struct four_doubles_t measurement = prof_measure(profile->measurer, NULL);

    if (profile->last_thread_data->fiber != thread_data->fiber)
    {
        switch_thread(profile, thread_data, measurement);
    }

    while (prof_frame_pop(thread_data->stack, measurement, profile->measurer))
    {
        ;
    }

    return ST_CONTINUE;
}

static void prof_stop_threads(prof_profile_t *profile) { rb_st_foreach(profile->threads_tbl, pop_frames, (st_data_t)profile); }

/* call-seq:
   new()

   Returns a new profiler.
*/
static VALUE prof_initialize(int argc, VALUE *argv, VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);

    profile->measurer = prof_get_measurer();
    profile->allow_exceptions = true;
    profile->merge_fibers = true;
    profile->timelines = rb_str_new2("");
    profile->headers = rb_str_new2("");
    profile->measurements = rb_str_new2("");
    profile->bf_version = rb_str_new2("");
    profile->context = rb_str_new2("");
    profile->title = rb_str_new2("");
    profile->probed_features = rb_str_new2("");
    profile->ruby_description = rb_str_new2("");
    profile->ruby_release_date = rb_str_new2("");
    profile->timeline_defined = false;
    profile->current_threshold = 0;
    profile->first_threshold = true;
    profile->threshold_initial_time = 0;
    profile->current_method = NULL;
    profile->timespan_threshold = 0;
    profile->timeline_rules = rb_ary_new();
    profile->timeline_rules_count = 0;
    profile->header_was_built = false;
    profile->last_event = RUBY_EVENT_END;

    assert(profile->include_threads_tbl == NULL);
    profile->include_threads_tbl = threads_table_create();
    VALUE thread = rb_thread_current();
    rb_st_insert(profile->include_threads_tbl, thread, Qtrue);
    prof_exclude_common_methods(self);

    return self;
}

/* call-seq:
   paused? -> boolean

   Returns whether a profile is currently paused.*/
static VALUE prof_paused(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->paused;
}

/* call-seq:
   running? -> boolean

   Returns whether a profile is currently running.*/
static VALUE prof_running(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->running;
}

/* call-seq:
   start -> self

   Starts recording profile data.*/
static VALUE prof_start(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);

    if (profile->header_was_built == false)
    {
        /* Headers */
        rb_str_cat2(profile->headers, "file-format: BlackfireProbe\nProbe-version: ");
        rb_str_cat2(profile->headers, StringValueCStr(profile->bf_version));
        rb_str_cat2(profile->headers, "\nContext: ");
        rb_str_cat2(profile->headers, StringValueCStr(profile->context));
        rb_str_cat2(profile->headers, "\nProfile-Title: ");
        rb_str_cat2(profile->headers, StringValueCStr(profile->title));
        rb_str_cat2(profile->headers, "\nCost-Dimensions: wt cpu al pmu\nProbed-Os: ");
        rb_str_cat2(profile->headers, RUBY_PLATFORM);
        rb_str_cat2(profile->headers, "\nProbed-Runtime: ");
        rb_str_cat2(profile->headers, StringValueCStr(profile->ruby_description));
        rb_str_cat2(profile->headers, " - ");
        rb_str_cat2(profile->headers, StringValueCStr(profile->ruby_release_date));
        rb_str_cat2(profile->headers, "\nProbed-Language: ruby\nProbed-Features: ");
        rb_str_cat2(profile->headers, StringValueCStr(profile->probed_features));
        rb_str_cat2(profile->headers, "\n");
        profile->header_was_built = true;
    }

    if (profile->running == Qtrue)
    {
        if (profile->paused == Qfalse)
        {
            rb_raise(rb_eRuntimeError, "Blackfire is not paused.");
        }
        else
        {
            profile->paused = Qfalse;
            profile->measurement_at_pause_resume = prof_measure(profile->measurer, NULL);
        }

        return self;
    }

    profile->running = Qtrue;
    profile->paused = Qfalse;
    profile->last_thread_data = threads_table_insert(profile, get_fiber(profile));

    prof_install_hook(self);
    return self;
}

/* call-seq:
   pause -> self

   Pauses collecting profile data. */
static VALUE prof_pause(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    if (profile->running == Qfalse)
    {
        rb_raise(rb_eRuntimeError, "Blackfire is not running.");
    }

    if (profile->paused == Qfalse)
    {
        profile->paused = Qtrue;
        profile->measurement_at_pause_resume = prof_measure(profile->measurer, NULL);
        rb_st_foreach(profile->threads_tbl, pause_thread, (st_data_t)profile);
    }

    return self;
}

/* call-seq:
   stop -> self

   Stops collecting profile data.*/
static VALUE prof_stop(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);

    if (profile->running == Qfalse)
    {
        rb_raise(rb_eRuntimeError, "Blackfire.start was not yet called");
    }

    prof_remove_hook(self);
    prof_stop_threads(profile);

    /* Unset the last_thread_data (very important!)
       and the threads table */
    profile->running = profile->paused = Qfalse;
    profile->last_thread_data = NULL;

    profile->current_method = NULL;

    return self;
}

/* call-seq:
   threads -> Array of Blackfire::Thread

   Returns an array of Blackfire::Thread instances that were profiled. */
static VALUE prof_threads(VALUE self)
{
    VALUE result = rb_ary_new();
    prof_profile_t *profile = prof_get_profile(self);
    rb_st_foreach(profile->threads_tbl, collect_threads, result);
    return result;
}

/* Document-method: Blackfire::Profile#Profile
   call-seq:
   profile(&block) -> self

   Profiles the specified block.

   profile = Blackfire::Profile.new
   profile.profile do
   ..
   end
*/
static VALUE prof_profile_object(VALUE self)
{
    int result;
    prof_profile_t *profile = prof_get_profile(self);

    if (!rb_block_given_p())
    {
        rb_raise(rb_eArgError, "A block must be provided to the profile method.");
    }

    prof_start(self);
    rb_protect(rb_yield, self, &result);
    self = prof_stop(self);

    if (profile->allow_exceptions && result != 0)
    {
        rb_jump_tag(result);
    }

    return self;
}

/* Document-method: Blackfire::Profile::Profile
   call-seq:
   profile(&block) -> Blackfire::Profile
   profile(options, &block) -> Blackfire::Profile

   Profiles the specified block and returns a Blackfire::Profile
   object. Arguments are passed to Profile initialize method.

   profile = Blackfire::Profile.profile do
   ..
   end
*/
static VALUE prof_profile_class(int argc, VALUE *argv, VALUE klass) { return prof_profile_object(rb_class_new_instance(argc, argv, cProfile)); }

/* call-seq:
   exclude_method!(module, method_name) -> self

   Excludes the method from profiling results.
*/
static VALUE prof_exclude_method(VALUE self, VALUE klass, VALUE msym)
{
    prof_profile_t *profile = prof_get_profile(self);

    if (profile->running == Qtrue)
    {
        rb_raise(rb_eRuntimeError, "Blackfire.start was already called");
    }

    st_data_t key = method_key(klass, msym);
    prof_method_t *method = method_table_lookup(profile->exclude_methods_tbl, key);

    if (!method)
    {
        method = prof_method_create(self, klass, msym, Qnil, 0);
        method_table_insert(profile->exclude_methods_tbl, method->key, method);
    }

    return self;
}

static VALUE prof_events(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->timelines;
}

static VALUE prof_header(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->headers;
}

static VALUE prof_bf_version(VALUE self, VALUE version)
{
    prof_profile_t *profile = prof_get_profile(self);
    profile->bf_version = version;
    return version;
}

static VALUE prof_context(VALUE self, VALUE ctx)
{
    prof_profile_t *profile = prof_get_profile(self);
    profile->context = ctx;
    return ctx;
}

static VALUE prof_title(VALUE self, VALUE title)
{
    prof_profile_t *profile = prof_get_profile(self);
    profile->title = title;
    return title;
}

static VALUE prof_probed_features(VALUE self, VALUE features)
{
    prof_profile_t *profile = prof_get_profile(self);
    profile->probed_features = features;
    return features;
}

static VALUE prof_ruby_description(VALUE self, VALUE description)
{
    prof_profile_t *profile = prof_get_profile(self);
    profile->ruby_description = description;
    return description;
}

static VALUE prof_ruby_release_date(VALUE self, VALUE release)
{
    prof_profile_t *profile = prof_get_profile(self);
    profile->ruby_release_date = release;
    return release;
}

static VALUE prof_timeline_defined(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    if (profile->timeline_defined == true)
    {
        return Qtrue;
    }
    else
    {
        return Qfalse;
    }
}

static VALUE prof_set_timeline_defined(VALUE self, VALUE timeline)
{
    prof_profile_t *profile = prof_get_profile(self);
    if (timeline == Qtrue)
    {
        profile->timeline_defined = true;
    }
    else
    {
        profile->timeline_defined = false;
    }

    return timeline;
}

static VALUE prof_set_timespan_threshold(VALUE self, VALUE threshold)
{
    prof_profile_t *profile = prof_get_profile(self);
    if (FIXNUM_P(threshold))
    {
        profile->timespan_threshold = FIX2LONG(threshold);
    }
    else
    {
        profile->timespan_threshold = 0;
    }
    return threshold;
}

static VALUE prof_add_timespan_rule(VALUE self, VALUE rule)
{
    prof_profile_t *profile = prof_get_profile(self);
    if (RB_TYPE_P(rule, T_STRING))
    {
        rb_ary_push(profile->timeline_rules, rule);
        profile->timeline_rules_count++;
    }
    return Qnil;
}

static VALUE prof_timespan_rules(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->timeline_rules;
}

static VALUE prof_timespan_rules_count(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return LONG2FIX(profile->timeline_rules_count);
}

static VALUE prof_parse_timespan(VALUE self, VALUE method, VALUE rule)
{
    prof_profile_t *profile = prof_get_profile(self);
    bool ret = false;
    if (RB_TYPE_P(method, T_STRING))
    {
        ret = parse_method_name_for_timeline(profile, NULL, method, FIX2LONG(rule));
    }
    return ret == true ? Qtrue : Qfalse;
}

void rp_init_profile(void)
{
    cProfile = rb_define_class_under(mProf, "Profile", rb_cObject);
    rb_define_alloc_func(cProfile, prof_allocate);

    rb_define_singleton_method(cProfile, "profile", prof_profile_class, -1);
    rb_define_method(cProfile, "initialize", prof_initialize, -1);
    rb_define_method(cProfile, "enable", prof_start, 0);
    rb_define_method(cProfile, "stop", prof_stop, 0);
    rb_define_method(cProfile, "disable", prof_pause, 0);
    rb_define_method(cProfile, "running?", prof_running, 0);
    rb_define_method(cProfile, "paused?", prof_paused, 0);
    rb_define_method(cProfile, "threads", prof_threads, 0);
    rb_define_method(cProfile, "exclude_method!", prof_exclude_method, 2);
    rb_define_method(cProfile, "profile", prof_profile_object, 0);

    /* protocol data */
    rb_define_method(cProfile, "events", prof_events, 0);
    rb_define_method(cProfile, "headers", prof_header, 0);
    rb_define_method(cProfile, "timeline_defined?", prof_timeline_defined, 0);

    /* setters */
    rb_define_method(cProfile, "bf_version=", prof_bf_version, 1);
    rb_define_method(cProfile, "context=", prof_context, 1);
    rb_define_method(cProfile, "title=", prof_title, 1);
    rb_define_method(cProfile, "probed_features=", prof_probed_features, 1);
    rb_define_method(cProfile, "ruby_description=", prof_ruby_description, 1);
    rb_define_method(cProfile, "ruby_release_date=", prof_ruby_release_date, 1);
    rb_define_method(cProfile, "timeline_defined=", prof_set_timeline_defined, 1);
    rb_define_method(cProfile, "timespan_threshold=", prof_set_timespan_threshold, 1);
    rb_define_method(cProfile, "add_timespan_rule", prof_add_timespan_rule, 1);
    rb_define_method(cProfile, "timespan_rules", prof_timespan_rules, 0);
    rb_define_method(cProfile, "timespan_rules_count", prof_timespan_rules_count, 0);
    rb_define_method(cProfile, "parse_for_timeline", prof_parse_timespan, 2);
}
