module Blackfire
  # The MethodInfo class is used to track information about each method that is profiled.
  # You cannot create a MethodInfo object directly, they are generated while running a profile.
  class MethodInfo
    include Comparable

    # Returns the full name of a class. The interpretation of method names is:
    #
    # * MyObject#test - An method defined in a class
    # * <Class:MyObject>#test - A method defined in a singleton class.
    # * <Module:MyObject>#test - A method defined in a singleton module.
    # * <Object:MyObject>#test - A method defined in a singleton object.
    def full_name
      decorated_class_name = case klass_flags
                             when 0x2
                               "<Class::#{klass_name}>"
                             when 0x4
                               "<Module::#{klass_name}>"
                             when 0x8
                               "<Object::#{klass_name}>"
                             else
                               klass_name
                             end

      "#{decorated_class_name}##{method_name}"
    end

    # The number of times this method was called
    def called
      measurement.called
    end

    # The time this method's children took to execute
    def children_time
      measurement.wall_time_total_time - measurement.wall_time_self_time - measurement.wall_time_wait_time
    end

    # :enddoc:
    # rubocop:disable Metrics/PerceivedComplexity
    def <=>(other)
      if other.nil?
        -1
      elsif full_name == other.full_name
        0
      elsif measurement.wall_time_total_time < other.measurement.wall_time_total_time
        -1
      elsif measurement.wall_time_total_time > other.measurement.wall_time_total_time
        1
      elsif call_trees.min_depth < other.call_trees.min_depth
        1
      elsif call_trees.min_depth > other.call_trees.min_depth
        -1
      else
        full_name <=> other.full_name
      end
    end
    # rubocop:enable Metrics/PerceivedComplexity

    def to_s
      "#{full_name} (c: #{called})"
    end
  end
end
