require "stringio"
require "socket"

module Blackfire
  class << self
    @current_connection = nil

    # Starts or resume profiling
    def enable(query = nil)
      logger = Logger.build
      logger.info "Start Blackfire Ruby Probe profiling"

      if defined?(@profile) && @profile
        return if @profile.running?

        @profile.enable
      else
        Query.parse(query)
        QueryTrace.cleanup
      end

      # basic settings
      @profile = Profile.new
      @profile.bf_version = Blackfire::VERSION
      @profile.context = __FILE__.split('/')[-1]
      @profile.title = Query.profile_title
      @profile.probed_features = Query.probed_features
      @profile.ruby_description = RUBY_DESCRIPTION
      @profile.ruby_release_date = RUBY_RELEASE_DATE
      @profile.timeline_defined = Query.flag_timespan == "1" ? true : false
      @profile.timespan_threshold = Query.timespan_threshold.to_i rescue 0

      begin
        open_connection
        return if @current_connection.nil?

        logger.info "Writting prolog"

        # Blackfire-Query header
        @current_connection.write "Blackfire-Query: #{Query.data}\n"

        # Blackfire-Probe header
        bf_probe_header = "Blackfire-Probe: "

        # "ruby" must be in lowercase to match the language in the metrics
        bf_probe_header += "ruby-#{RUBY_VERSION}-p#{RUBY_PATCHLEVEL}"
        bf_probe_header += ", timespan" if @profile.timeline_defined?
        bf_probe_header += ", blackfire_yml" if BlackfireYAML.defined?
        bf_probe_header += "\n"

        @current_connection.write bf_probe_header

        # Blackfire-Auth header
        @current_connection.write "Blackfire-Auth: #{Config.server_id}:#{Config.server_token}\n" if Config.valid_auth_header?

        # send
        @current_connection.write "\n"

        logger.info "Waiting agent response"
        # wait data
        answer = @current_connection.gets

        if answer.start_with? "Blackfire-Response".freeze
          Rack::BlackfireRack.define_header(answer)
          BlackfireYAML.send_file(@current_connection) if BlackfireYAML.must_be_sent? answer

          # Check for the Blackfire-Keys header
          header = @current_connection.gets
          if header.start_with? "Blackfire-Keys".freeze
            # FIXME: store the keys
            header = @current_connection.gets
          end

          FnArgs.cleanup
          loop do
            break if header == "\n"

            @profile.add_timespan_rule(header.gsub("Blackfire-Timespan: ", "")) if header.include? "Blackfire-Timespan"
            FnArgs.add_new_rule(header) if FnArgsRule.must_be_added? header
            header = @current_connection.gets
          end
        else
          logger.error answer
          @current_connection.close
          return
        end
      rescue StandardError => e
        logger.info "Error while trying to talk with socket #{e.inspect}"
      end

      # start must be last line at the enable

      @profile.enable

      @profile
    end

    # Stops profiling
    # rubocop:disable Metrics/CyclomaticComplexity
    # rubocop:disable Metrics/PerceivedComplexity
    # rubocop:disable Metrics/MethodLength
    def end
      logger = Logger.build
      logger.info "Stop called, ending profiling"

      return unless check_running?

      result = @profile.stop
      @profile = nil

      begin
        logger.info "Writing profile"
        printer = Blackfire::FlatPrinter.new(result)
        io = StringIO.new
        printer.print(io)
        logger.debug io.string
        # send profile
        @current_connection.write result.headers +
                                  io.string.split("\n")[0] +
                                  "\n" +
                                  result.events +
                                  "\n" +
                                  io.string.gsub("#{io.string.split("\n")[0]}\n", "")

        logger.info "Closing connection"
        # ends connections
        @current_connection.close
      rescue StandardError => e
        logger.info "Error while trying to talk with socket #{e.inspect}"
      end

      result
    end
    # rubocop:enable Metrics/CyclomaticComplexity
    # rubocop:enable Metrics/PerceivedComplexity
    # rubocop:enable Metrics/MethodLength

    def check_running?
      return false unless defined?(@profile) && @profile

      @profile.running?
    end

    private

    def open_connection
      blackfire_socket = Config.agent_socket
      return if blackfire_socket.empty?

      unless @current_connection.nil?
        # begin
        @current_connection.close
        @current_connection = nil
        # rescue

        # end
      end

      if Config.unix_socket?
        data = blackfire_socket.split("://")[1]
        @current_connection = UNIXSocket.new(data)
      else
        data = blackfire_socket.split("//")[1].split(":")
        server = data[0]
        port = data[1]
        @current_connection = TCPSocket.new(server, port)
      end
    end
  end
end
